/*
Copyright (c) 1993, Clemens Wagner
 
Permission to use, copy, modify, and distribute this software and 
its documentation for any purpose is hereby granted without fee, provided
that
(i) you have alays to distribute the whole archive
(ii) the above copyright notices and this permission notice appear in
all copies of the software and related documentation, changed or new
code has to be marked
(iii) else, there are the same conditions as for InterViews

THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 */

#include <Graphic/graphicmaster.h>
#include <Graphic/glyphviewer.h>
#include <Graphic/handles.h>

#include "grabber.bm"
#include "grabberMask.bm"

#include <InterViews/bitmap.h>
#include <InterViews/brush.h>
#include <InterViews/canvas.h>
#include <InterViews/character.h>
#include <InterViews/color.h>
#include <InterViews/cursor.h>
#include <InterViews/display.h>
#include <InterViews/font.h>
#include <InterViews/handler.h>
#include <InterViews/hit.h>
#include <InterViews/layout.h>
#include <InterViews/tformsetter.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>
#include <IV-X11/xcanvas.h>
#include <OS/list.h>
#include <OS/math.h>
#include <OS/memory.h>
#include <OS/string.h>
#include <stdlib.h>
#include <stream.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

//
//
//

implementPtrList(GraphicList,Graphic)

static Cursor *grabber_cursor=nil;
static Cursor *window_cursor=nil;

GraphicMaster::GraphicMaster(GlyphViewer *gv,Graphic *gr,const Color *bg) : PolyGraphic(gr)
{
	_gv=gv;
	Resource::ref(gv);
	_gr_list=new GraphicList;
	if(grabber_cursor==nil)
	{
		Bitmap *grabber=new Bitmap
		(
			grabber_bits,grabber_width,grabber_height,
			grabber_x_hot,grabber_y_hot
		);
		Bitmap *grabber_mask=new Bitmap
		(
			grabberMask_bits,grabberMask_width,grabberMask_height,
			grabberMask_x_hot,grabberMask_y_hot
		);
		grabber_cursor=new Cursor(grabber,grabber_mask);
	}
	_bg=bg;
	Resource::ref(_bg);
}

GraphicMaster::~GraphicMaster()
{
	delete _gr_list;
	Resource::unref(_bg);
	Resource::unref(_gv);
}

void GraphicMaster::background(const Color *bg)
{
	if(bg!=nil)
		Resource::ref(bg);
	if(_bg!=nil)
		Resource::unref(_bg);
	_bg=bg;
}

Glyph *GraphicMaster::clone() const
{
	Graphic *pg=new GraphicMaster(glyphviewer(),(Graphic *) this);
	GlyphIndex i,count=_body->count();

	for(i=0;i<count;i++)
	{
		Glyph *gr=_body->component(i);
		pg->append(gr->clone());
	}
	return pg;
}

void GraphicMaster::request(Requisition &req) const
{
	PolyGraphic::request(req);
	Requirement &rx=req.x_requirement();
	rx.alignment(0.0);
	Requirement &ry=req.y_requirement();
	ry.alignment(0.0);
}

void GraphicMaster::allocate(Canvas *c,const Allocation &a,Extension &ext)
{
	if(!_a.equals(a,0.001))
	{
		/*
		Allocation b(a);
		Allotment &ax=_a.x_allotment();
		Allotment &ay=_a.y_allotment();
		Allotment &bx=b.x_allotment();
		Allotment &by=b.y_allotment();
		//translate((bx.span()-ax.span())/2.0,(by.span()-ay.span())/2.0);
		//translate(bx.begin()-ax.begin(),by.begin()-ay.begin());
		*/
		/*
		Coord x,y;
		Coord a00,a01,a10,a11,a20,a21;
		Transformer *t=transformer();

		_a=a;
		if(t==nil)
			a20=a21=0.0;
		else
			t->matrix(a00,a01,a10,a11,a20,a21);
		centerpoint(x,y);
		translate(x,y);
		*/
		_a=a;
		_gv->commit_adjustment(Dimension_X);
		_gv->commit_adjustment(Dimension_Y);
	}
	if(c!=nil)
		PolyGraphic::allocate(c,a,ext);
}

void GraphicMaster::update(GlyphViewer *gv)
{
	GlyphIndex i,count=count_();

	Graphic::update();
	for(i=0;i<count;i++)
		component_(i)->update(gv);
}

void GraphicMaster::get_sbounds(Coord &l,Coord &b,Coord &r,Coord &t) const
{
	long count=_gr_list->count();

	if(count>0)
	{
		Coord ll,bb,rr,tt;
		long i;

		_gr_list->item(0)->getbounds(l,b,r,t);
		for(i=1;i<count;i++)
		{
			_gr_list->item(i)->getbounds(ll,bb,rr,tt);
			l=Math::min(l,ll);
			b=Math::min(b,bb);
			r=Math::max(r,rr);
			t=Math::max(t,tt);
		}
	}
}

boolean GraphicMaster::is_selected(const Graphic *gr) const
{
	long l,count=_gr_list->count();

	for(l=0;l<count;l++)
		if(gr==_gr_list->item(l))
			return true;
	return false;
}

boolean GraphicMaster::select(Graphic *gr,Canvas *c,boolean always)
{
	long l,count=_gr_list->count();
	boolean found=false;

	for(l=0;l<count;l++)
	{
		if(gr==_gr_list->item(l))
		{
			found=true;
			break;
		}
	}
	if(found)
	{
		if(!always)
		{
			_gr_list->remove(l);
			gr->delete_handles();
			return false;
		}
	}
	else
	{
		_gr_list->append(gr);
		gr->create_handles(c);
	}
	return true;
}

void GraphicMaster::unselect(Graphic *gr)
{
	long l,count=_gr_list->count();

	for(l=0;l<count;l++)
	{
		if(gr==_gr_list->item(l))
		{
			gr->delete_handles();
			_gr_list->remove(l);
			return;
		}
	}
}

void GraphicMaster::clear_selection()
{
	long l,count=_gr_list->count();

	for(l=count-1;l>=0;l--)
	{
		_gr_list->item(l)->delete_handles();
		_gr_list->remove(l);
	}
}

void GraphicMaster::transform_selection(const Transformer *tf)
{
	Coord l,b,r,t;
	long i,count=_gr_list->count();
	GlyphViewer *gv=glyphviewer();

	if(gv!=nil)
	{
		for(i=0;i<count;i++)
		{
			Graphic *gr=_gr_list->item(i);
			Transformer *gt=gr->transformer();
	
			gr->delete_handles();
			gr->getbounds(l,b,r,t);
			l-=Handles::tol();
			b-=Handles::tol();
			r+=Handles::tol();
			t+=Handles::tol();
			gv->damage(l,b,r,t);
			if(gt==nil)
			{
				gt=new Transformer(tf);

				gr->transformer(gt);
				gt->unref();
			}
			else
				gt->postmultiply(*tf);
			gr->create_handles();
			gr->getbounds(l,b,r,t);
			l-=Handles::tol();
			b-=Handles::tol();
			r+=Handles::tol();
			t+=Handles::tol();
			gv->damage(l,b,r,t);
		}
		update(gv);
		gv->repair();
	}
}

void GraphicMaster::glyphviewer(GlyphViewer *gv)
{
	Resource::ref(gv);
	Resource::unref(_gv);
	_gv=gv;
}

void GraphicMaster::centerpoint(Coord &x,Coord &y) const
{
	Coord l,b,r,t;

	l=_a.left();
	b=_a.bottom();
	r=_a.right();
	t=_a.top();
	x=(l+r)/2.0;
	y=(b+t)/2.0;
}

ClassId GraphicMaster::get_classid() const
{
	return GRAPHIC_MASTER;
}

boolean GraphicMaster::is_a(ClassId id) const
{
	return id==GRAPHIC_MASTER || PolyGraphic::is_a(id);
}

//

void GraphicMaster::draw_gs(Canvas *c,Graphic *gs)
{
	GraphicList *s=selection();
	long i,count=s->count();

	//draw_frame(c);
	PolyGraphic::draw_gs(c,gs);
	for(i=0;i<count;i++)
		s->item(i)->draw_handles(c);
}

void GraphicMaster::drawclipped_gs(Canvas *c,Coord l,Coord b,Coord r,Coord t,Graphic *gs)
{
	Coord tol=Handles::tol();
	GraphicList *s=selection();
	long i,count=s->count();

	c->push_clipping();
	c->clip_rect(_a.left(),_a.bottom(),_a.right(),_a.top());
	if (_bg != nil)
		c->fill_rect(l,b,r,t,_bg);
	//draw_frame(c);
	PolyGraphic::drawclipped_gs(c,l,b,r,t,gs);
	for(i=0;i<count;i++)
		s->item(i)->draw_handles(c);
	c->pop_clipping();
}

//

void GraphicMaster::draw_frame(Canvas *c)
{
	Brush *b=new Brush(1.0);
	Color *col=new Color(0.0,0.0,0.0,1.0);
	Transformer *t=transformer();

	if(t!=nil)
	{
		c->push_transform();
		c->transform(t);
	}
	c->rect(_gv->lower(Dimension_X),_gv->lower(Dimension_Y),_gv->upper(Dimension_X),_gv->upper(Dimension_Y),col,b);
	if(t!=nil)
		c->pop_transform();
}
